import random
import timeit
import datetime
import time
import os
import math
from optparse import OptionParser
import shutil
	

class CplexBinaryEncodingGenerator:
	
	def __init__(self):
		self.variableCounter = 0;
		self.constraintCounter = 0;
		self.allBinaryVars = []

	def getaFileobject(self, outputFilePath):
		filename = outputFilePath
		if os.path.exists(filename):
			os.remove(filename)
		fo = open(filename, "a")
		return fo
	
	def closeFileobject(self, fo):
		fo.close()
		
	def getCostMatrixFromFile(self, costMatrixFilePath):
		costMatrix = []
		self.numberOfNodes = 0;
		self.euclideanMaximumCost = 0;
		with open(costMatrixFilePath) as infile:
			for line in infile:			
				costs = line.split(',')
				if(len(costs) == 1):
					self.numberOfNodes = int(costs[0])
					continue
				intCosts = []
				for c in costs:
					cost = int(c)
					intCosts.append(int(c))
					if( cost > self.euclideanMaximumCost):
						self.euclideanMaximumCost = cost						
				costMatrix.append(intCosts)
				
		return costMatrix
	
	def getEdgekey(self, i, j):
		return "edge(" + str(i) + "," + str(j) + ")"
	
	def getNewVar(self):
		self.variableCounter += 1
		newvar = "x" + str(self.variableCounter)
		self.allBinaryVars.append(newvar)
		return newvar
	
	def generateObjectiveFunction(self, costMatrix):
		string = "";
		self.edge_variable_map = {}
	
		edgeVarMapFilePath = self.setFolderPath + "/" + "EdgeVarMapFile.map";
		if(os.path.exists(edgeVarMapFilePath)):
			os.remove(edgeVarMapFilePath)		
		edgeVarMapFile = self.getaFileobject(edgeVarMapFilePath)
		
		
		for i in range(1, len(costMatrix) + 1):
			row = costMatrix[i-1]
			for j in range(1, len(row) + 1):
				if (i == j): continue
				edgeKey = self.getEdgekey(i,j)
				edgeVar = self.getNewVar()
				self.edge_variable_map[edgeKey] = edgeVar
				cost = str(row[j-1])
				string += "+" + cost + " " + edgeVar + " "
				evmStr = edgeKey + " : " + self.edge_variable_map[edgeKey] + "\n"
				edgeVarMapFile.write(evmStr)
			if(i != len(costMatrix)):
				string += "\n"
		
		edgeVarMapFile.close()		
		return string
	
	def inOutDegreeConstraints(self, numberOfNodes):
		PBString = ""		  
		#"Out going edge from i"
		for i in range(1, numberOfNodes + 1):
			for j in range(1, numberOfNodes + 1):				
				if(i != j):
					edgekey = self.getEdgekey(i,j)
					edgeVar = self.edge_variable_map[edgekey]
					PBString += "+" + "1 " + edgeVar + " "
			PBString += "= 1 \n"
			self.constraintCounter += 1;

		
		PBString += "\n"
		#"In coming edge at j"
		for j in range(1, numberOfNodes + 1):
			for i in range(1, numberOfNodes + 1):
				if(i != j):
					edgekey = self.getEdgekey(i,j)
					edgeVar = self.edge_variable_map[edgekey]
					PBString += "+" + "1 " + edgeVar + " " 
			PBString += "= 1 \n"
			self.constraintCounter += 1;
		return PBString
	
	def getUvariable(self, nBits, sign = "+"):
		u = ""
		for i in range(0, len(nBits)):
			d = int(math.pow( 2, i ))
			u += sign + str(d) + " " + nBits[i] + " "
		return u
	
	def getNbitsVar(self, n):
		var = []
		for i in range(0, n ):
			v = self.getNewVar()
			var.append(v)
		return var
	
	def subToursEliminationConstraints(self, totalNodes):
		s ="\n"
		self.uVar_bits_maps = {}
		bitCount = len(bin(totalNodes)[2:])
		
		# creates ui (i = 2..n)
		for i in range(2, totalNodes + 1):
			uiVarkey = "u" + str(i)
			uiBits = self.getNbitsVar(bitCount)
			self.uVar_bits_maps[uiVarkey] = uiBits
			#print uiBits
			
		#Eq 11
		for i in range(2, totalNodes + 1):
			
			uiVarkey = "u" + str(i)
			uiBits = self.uVar_bits_maps[uiVarkey] 
			ui = self.getUvariable(uiBits)
			
			for j in range(2, totalNodes + 1):
				if(i == j ): continue
				
				ujVarkey = "u" + str(j)
				ujBits = self.uVar_bits_maps[ujVarkey]
				negUj = self.getUvariable(ujBits,"-")
				
				x_ij_varKey = self.getEdgekey(i,j)
				x_ij_var = self.edge_variable_map[x_ij_varKey]
				
				x_ji_varKey = self.getEdgekey(j,i)
				x_ji_var = self.edge_variable_map[x_ji_varKey]
				
				c = "{0}{1} +{2} {3} +{4} {5} <= {6}\n".format(ui , negUj , totalNodes - 1, x_ij_var, totalNodes -3, x_ji_var, totalNodes -2)
				s += c
				self.constraintCounter += 1
			s += "\n"
		#Eq 12
		tempVar = self.getNewVar()
		for i in range(2, totalNodes + 1):
			
			uiVarkey = "u" + str(i)
			uiBits = self.uVar_bits_maps[uiVarkey]
			ui = self.getUvariable(uiBits)
			negUi = self.getUvariable(uiBits,"-")
			
			x_i1_varKey = self.getEdgekey(i,1)			
			x_i1_var = self.edge_variable_map[x_i1_varKey] 
			
			x_1i_varKey = self.getEdgekey(1,i)
			x_1i_var = self.edge_variable_map[x_1i_varKey] 
			
			sum_xij_vars = "";
			sum_xji_vars = "";
			
			for j in range(2, totalNodes + 1):
				if(i == j ): continue
				
				x_ij_varKey = self.getEdgekey(i,j)
				x_ij_var = self.edge_variable_map[x_ij_varKey]
				sum_xij_vars = sum_xij_vars + " +1 " + x_ij_var
				
				x_ji_varKey = self.getEdgekey(j,i)
				x_ji_var = self.edge_variable_map[x_ji_varKey]			
				sum_xji_vars = sum_xji_vars + " +1 " + x_ji_var
			
			c1 = "+1 " + tempVar + " +" + str(totalNodes-3) + " " + x_i1_var + " " + sum_xji_vars + " " + negUi + " <= 0 "
			self.constraintCounter += 1
			#c1 = "+ 1 tempVar + {0} {1} {2} {} <= 0".format(totalNodes-3, x_i1_var, sum_xji_vars , negUi)
			c2 = ui + "-" + str(totalNodes - 1) + " " + tempVar + " +" + str(totalNodes- 3) + " " + x_1i_var + " " + sum_xij_vars + " <= 0 "
			self.constraintCounter += 1
			#c2 = "{0} - {1} tempVar + {2} {3} {4} <= 0".format(ui_var, totalNodes - 1, totalNodes- 3, x_1i_var, sum_xij_vars)
			s += c1 + "\n" + c2
			s += "\n\n"
			
		s += "+1 " + tempVar + " = 1 ;";
		self.constraintCounter += 1
		#print s		
		return s
	
	def subToursEliminationConstraints_old(self, totalNodes):
		
		subTourConstraints = "\n"
		boundConstraints = "\n"	
		self.uVar_bits_maps = {}
		bitCount = len(bin(totalNodes)[2:])
		
		for i in range(2, totalNodes + 1):
			
			varkey = "u" + str(i)
			#subTourConstraints += "*" + varkey + "\n"
			uiBits = self.getNbitsVar(bitCount)
			self.uVar_bits_maps[varkey] = uiBits
			ui = self.getUvariable(uiBits)
			
			uiRangeConstraint =  ui + " >= 2 \n"
			boundConstraints +=  uiRangeConstraint
			self.constraintCounter += 1
			
			uiRangeConstraint =  ui + " <= {0} \n".format(totalNodes)
			boundConstraints +=  uiRangeConstraint
			self.constraintCounter += 1
			
			boundConstraints += "\n"
		
		for i in range(2, totalNodes + 1):
			uiVarkey = "u" + str(i)
			uiBits = self.uVar_bits_maps[uiVarkey]
			ui = self.getUvariable(uiBits)
				   
			for j  in range(2, totalNodes + 1):
				if (i == j): continue		
				
				ujVarkey = "u" + str(j)
				ujBits = self.uVar_bits_maps[ujVarkey]
				negUj = self.getUvariable(ujBits,"-")
								
				edgekey = self.getEdgekey(i,j)
				varName = self.edge_variable_map[edgekey]				
				
				p = totalNodes -1
				# ui - uj + p xij <= p-1
				#subTourConstraints += "*u{0} - u{1} + {2} {3} <= {4}\n".format(i,j,p,varName,p - 1)	
				subTourConstraints +=  ui + negUj + "+{0} {1} <= {2} \n".format(p,varName,p-1)
				self.constraintCounter += 1
			
			subTourConstraints += "\n"
				
		return (subTourConstraints, boundConstraints)
	
	def getAllBinaryVars(self):
		string = ""
		for binaryVar in self.allBinaryVars:
			string += binaryVar + "\n"
		return string
	
	def generateEncoding(self,totalNodes):
		
		self.variableCounter = 0;
		self.constraintCounter = 0;
		self.allBinaryVars = []
		
		#(costMatrix, numberOfNodes) = self.getCostMatrixFromFile(options.costMatrixFilePath)
		costMatrix = self.costMatrix
		numberOfNodes = totalNodes
		
		objectiveFunction = self.generateObjectiveFunction(costMatrix)
		in_out_edge_contraint = self.inOutDegreeConstraints(numberOfNodes)
		
		PBstring = "Minimize \n";
		PBstring += objectiveFunction + "\n"
		PBstring += "\nSubject To\n" + in_out_edge_contraint
		subTourConstraints = self.subToursEliminationConstraints(numberOfNodes)
		PBstring += subTourConstraints
		#PBstring += boundConstraints
		binaryVars =  "\nBinary\n" + self.getAllBinaryVars()
		PBstring +=  binaryVars
		PBstring +=  "\nEnd"
		PBstring = PBstring
		return PBstring

	def optionParse(self):
		parser = OptionParser()
		parser.add_option("--costMatrixFilePath", dest="costMatrixFilePath", help="Cost Matrix File path", type="string", default  = 'EuclideanDistanceCostmatrix.txt' )
		parser.add_option("--outputFilePath", dest="outputFilePath", help="output file path", type="string" , default  = 'PBEncoding_From_EuclideanDistanceCostmatrix.opb')
		return parser.parse_args()
	
	def creatFolder(self, path):
		if(os.path.exists(path)):
			shutil.rmtree(path)
		os.makedirs(path)
	
	def deleteThanCreateFolder(self, path):
		if(os.path.exists(path)):
			shutil.rmtree(path)
		os.makedirs(path)
	
	def createIfNotExists(self, path):
		if(os.path.exists(path) == False):
			os.makedirs(path)
			
	def generate(self):
		
		self.benchMarkFolder = "../../../BenchMarks";
		costFilesPath = ['EuclideanCost.txt','OceanDataCost.txt','RandomCost.txt',
				'noisy_EuclideanCost/RandomSymmetric_EuclideanCostModified.txt',
				'noisy_EuclideanCost/AsymmetricRandomNoisy_50PercentEuclideanCostModified.txt',
				'noisy_EuclideanCost/AsymmetricNoisy_50PercentEuclideanCostModified.txt',
				'noisy_EuclideanCost/SymmetricNoisy_50PercentEuclideanCostModified.txt'
				]
		
		#costFilesPath = ['OceanDataCost.txt']
		
		nodeFolders  = os.listdir(self.benchMarkFolder)
		for nodeFolderName in nodeFolders:
			print "-------------"
			#if(nodeFolderName != "5_Nodes"): continue;
			craftedFolderPath = self.benchMarkFolder + "/" + nodeFolderName
			if(os.path.isdir(craftedFolderPath)):
				print nodeFolderName
				setFolders  = os.listdir(craftedFolderPath)
				for setFolder in setFolders:
					#if(setFolder != "baseTest_set"): continue;
					
					setFolderPath = craftedFolderPath + "/" + setFolder
					self.setFolderPath = setFolderPath
					
					encodingFolderPath = setFolderPath + "/encodings"
					self.createIfNotExists(encodingFolderPath)
					
					cplexEncodingFolderPath = setFolderPath + "/encodings/cplex"
					self.createIfNotExists(cplexEncodingFolderPath)
					
					cplexEncodingFolderPath = setFolderPath + "/encodings/cplex/binary"
					self.createIfNotExists(cplexEncodingFolderPath)
					
					cplexDLencodingFolderPath = setFolderPath + "/encodings/cplex/binary/dlEncoding"
					self.deleteThanCreateFolder(cplexDLencodingFolderPath)
					
					costFilespath = setFolderPath + "/costFiles/";
					for costfile in costFilesPath:
						costFilePath = costFilespath + costfile
						if(os.path.exists(costFilePath)):
							self.costMatrix = self.getCostMatrixFromFile(costFilePath)
							#print costFilePath
							#self.printCostMatrix(self.costMatrix)
							encoding = self.generateEncoding(self.numberOfNodes)
							costFilename =  os.path.basename(costFilePath).split('.')[0]
							outputFilePath = cplexDLencodingFolderPath + "/cplex_binary_dl_" + costFilename + ".lp"
							#print outputFilePath
							fileObj = open(outputFilePath,"w+")
							fileObj.write(encoding)
							fileObj.close()
						else:
							print "Not Exist: {0}".format(costFilePath)
				
				print "-------------"

if __name__ == '__main__':
	
	generator = CplexBinaryEncodingGenerator()
	generator.generate()
	
	

