import random
import timeit
import datetime
import time
import os
import math
from optparse import OptionParser
import shutil
	

class NewPBencodingGenerator:
	
	def __init__(self):
		self.variableCounter = 0;
		self.constraintCounter = 0;

	def getaFileobject(self, outputFilePath):
		filename = outputFilePath
		if os.path.exists(filename):
			os.remove(filename)
		fo = open(filename, "a")
		return fo
	
	def closeFileobject(self, fo):
		fo.close()
		
	def getCostMatrixFromFile(self, costMatrixFilePath):
		costMatrix = []
		numberOfNodes = 0;
		with open(costMatrixFilePath) as infile:
			for line in infile:			
				costs = line.split(',')
				if(len(costs) == 1):
					numberOfNodes = int(costs[0])
					continue
				intCosts = []
				for c in costs:
					cost = int(c)
					intCosts.append(int(c))
				costMatrix.append(intCosts)
				
		return (costMatrix, numberOfNodes)
	
	def getEdgekey(self, i, j):
		return "edge(" + str(i) + "," + str(j) + ")"
		
	def generateObjectiveFunction(self, costMatrix, options):
		string = "";
		self.edge_variable_map = {}
		
		edgeVarMapFilePath = options.absFolderPath + "/" + "EdgeVarMapFile.map";
		if(os.path.exists(edgeVarMapFilePath)):
			os.remove(edgeVarMapFilePath)		
		edgeVarMapFile = self.getaFileobject(edgeVarMapFilePath)

		for i in range(1, len(costMatrix) + 1):
			row = costMatrix[i-1]
			for j in range(1, len(row) + 1):
				if (i == j): continue
				edgeKey = self.getEdgekey(i,j)
				self.variableCounter += 1 
				edgeVar = "x" + str(self.variableCounter)
				self.edge_variable_map[edgeKey] = edgeVar
				cost = str(row[j-1])
				string += "+" + cost + " " + edgeVar + " "
				evmStr = edgeKey + " : " + self.edge_variable_map[edgeKey] + "\n"
				edgeVarMapFile.write(evmStr)
			if(i != len(costMatrix)):
				string += "\n"
		
		edgeVarMapFile.close()
		
		return string
	
	def inOutDegreeConstraints(self, numberOfNodes):
		PBString = ""		  
		#"Out going edge from i"
		for i in range(1, numberOfNodes + 1):
			for j in range(1, numberOfNodes + 1):				
				if(i != j):
					edgekey = self.getEdgekey(i,j)
					edgeVar = self.edge_variable_map[edgekey]
					PBString += "+" + "1 " + edgeVar + " "
			PBString += "= 1 ;\n"
			self.constraintCounter += 1;

		
		PBString += "\n"
		#"In coming edge at j"
		for j in range(1, numberOfNodes + 1):
			for i in range(1, numberOfNodes + 1):
				if(i != j):
					edgekey = self.getEdgekey(i,j)
					edgeVar = self.edge_variable_map[edgekey]
					PBString += "+" + "1 " + edgeVar + " " 
			PBString += "= 1 ;\n"
			self.constraintCounter += 1;
		return PBString
	
	def getUvariable(self, nBits, sign = "+"):
		u = ""
		for i in range(0, len(nBits)):
			d = int(math.pow( 2, i ))
			u += sign + str(d) + " " + nBits[i] + " "
		return u
	
	def getNbitsVar(self, n):
		var = []
		for i in range(0, n ):
			self.variableCounter += 1
			var.append("x%d" % self.variableCounter)
		return var
	
	def subToursEliminationConstraints(self, totalNodes):
		
		subTourConstraints = "\n"		
		self.uVar_bits_maps = {}
		bitCount = len(bin(totalNodes)[2:])
		
		self.variableCounter += 1
		commonVar = "x{0}".format(self.variableCounter)
		subTourConstraints += "+1 {0} = 1;\n".format(commonVar)
		
		for i in range(2, totalNodes + 1):
			
			varkey = "u" + str(i)
			subTourConstraints += "*" + varkey + "\n"
			uiBits = self.getNbitsVar(bitCount)
			self.uVar_bits_maps[varkey] = uiBits
			ui = self.getUvariable(uiBits)
			
			uiRangeConstraint =  ui + " >= 2 ;\n"
			subTourConstraints +=  uiRangeConstraint
			self.constraintCounter += 1
			
			negUi = self.getUvariable(uiBits,"-")
			var = "+{0} {1} ".format(totalNodes, commonVar) 
			uiRangeConstraint = var + negUi + " >= 0;\n"
			subTourConstraints +=  uiRangeConstraint
			self.constraintCounter += 1
			
			subTourConstraints += "\n"
		
		for i in range(2, totalNodes + 1):
			uiVarkey = "u" + str(i)
			uiBits = self.uVar_bits_maps[uiVarkey]
			negUi = self.getUvariable(uiBits, "-")
				   
			for j  in range(2, totalNodes + 1):
				if (i == j): continue		
				
				ujVarkey = "u" + str(j)
				ujBits = self.uVar_bits_maps[ujVarkey]
				uj = self.getUvariable(ujBits)
								
				edgekey = self.getEdgekey(i,j)
				varName = self.edge_variable_map[edgekey]				
				
				p = totalNodes -1
				# ui - uj + p xij <= p-1
				subTourConstraints += "*u{0} - u{1} + {2} {3} <= {4}\n".format(i,j,p,varName,p - 1)
				var = "+{0} {1} ".format(p-1, commonVar)
				subTourConstraints +=  var + negUi + uj + "-{0} {1} >= 0; \n".format(p,varName)
				self.constraintCounter += 1
			
			subTourConstraints += "\n"
		
		return subTourConstraints
		
		'''
		s ="\n"
		for i in range(2, numberOfNodes + 1):
			for j in range(2, numberOfNodes + 1):
				if(i == j ): 
					continue
				varName = "x{0}_{1}".format(i,j)
				p = numberOfNodes -1
				# ui - uj + p xij <= p-1
				c = "u{0} - u{1} + {2} {3} <= {4}\n".format(i,j,p,varName,p - 1)
				s += c
			s += "\n"
		return s
		'''
	
	def generate(self,options):
		
		self.variableCounter = 0;
		self.constraintCounter = 0;
		
		(costMatrix, numberOfNodes) = self.getCostMatrixFromFile(options.costMatrixFilePath)
		
		outPutfile = self.getaFileobject(options.outputFilePath)
		
		objectiveFunction = self.generateObjectiveFunction(costMatrix, options)
		in_out_edge_contraint = self.inOutDegreeConstraints(numberOfNodes)
		
		PBstring = "min: \n";
		PBstring += objectiveFunction + ";\n"
		PBstring += "\n" + in_out_edge_contraint
		subtourEliminationConstraints = self.subToursEliminationConstraints(numberOfNodes)
		PBstring += subtourEliminationConstraints
		
		fileHeader = "* #variable= {0} #constraint= {1}\n".format(self.variableCounter, self.constraintCounter)
		PBstring = fileHeader + PBstring
		
		outPutfile.write(PBstring)
		self.closeFileobject(outPutfile)
		
		#print objectiveFunction

	def optionParse(self):
		parser = OptionParser()
		parser.add_option("--costMatrixFilePath", dest="costMatrixFilePath", help="Cost Matrix File path", type="string", default  = 'EuclideanDistanceCostmatrix.txt' )
		parser.add_option("--outputFilePath", dest="outputFilePath", help="output file path", type="string" , default  = 'PBEncoding_From_EuclideanDistanceCostmatrix.opb')
		return parser.parse_args()
	
	def creatEncodingFolder(self, absFolderPath):
		encodingFolder = absFolderPath + "/encodings"
		if(os.path.exists(encodingFolder)):
			shutil.rmtree(encodingFolder)
		os.makedirs(encodingFolder)

	
	def creatFolder(self, path):
		if(os.path.exists(path)):
			shutil.rmtree(path)
		os.makedirs(path)

if __name__ == '__main__':
	
	generator = NewPBencodingGenerator()	
	(options, args) = generator.optionParse()
	
	print "------"
	benchMarkFolder = "../BenchMarks";
	costFilesPath = ['costFiles/EuclideanCost.txt','costFiles/OceanDataCost.txt','costFiles/RandomCost.txt',
				'costFiles/noisy_EuclideanCost/RandomSymmetric_EuclideanCostModified.txt',
				'costFiles/noisy_EuclideanCost/AsymmetricRandomNoisy_50PercentEuclideanCostModified.txt',
				'costFiles/noisy_EuclideanCost/AsymmetricNoisy_50PercentEuclideanCostModified.txt',
				'costFiles/noisy_EuclideanCost/SymmetricNoisy_50PercentEuclideanCostModified.txt'
				]
	#costFiles = ['costFiles/EuclideanCost.txt']
	nodeFoldes = os.listdir(benchMarkFolder)	
	for nodeFolderName in nodeFoldes:
		#print folderName
		print "Generating encoding for {0} folder".format(nodeFolderName)
		#if(str(nodeFolderName) != "4_Nodes"): continue;
		craftedFolderPath = benchMarkFolder + "/" + nodeFolderName		
		setFoldes = os.listdir(craftedFolderPath)
		for setFolder in setFoldes:
			setFolderPath = craftedFolderPath + '/' + setFolder			
			encodingFolderPath = setFolderPath + "/encodings/clasp"
			generator.creatFolder(encodingFolderPath)
			options.absFolderPath = setFolderPath
			for relCostFilePath in costFilesPath:
				costFilePath = setFolderPath + "/" + relCostFilePath
				if(os.path.exists(costFilePath)):					
					options.costMatrixFilePath = costFilePath;					
					costFilename =  os.path.basename(options.costMatrixFilePath).split('.')[0]					
					options.outputFilePath = encodingFolderPath + "/PBencd_" + costFilename + ".opb"
					if(os.path.exists(options.outputFilePath)):
						os.remove(options.outputFilePath)
					generator.generate(options)
				else:
					print "no file found {0}.".format(costFilePath)
		print "------"
	

