import time
import os
import math
import random
import shutil

class CplexBenchmarkGen:
	def __init__(self):
		self.varMap = 0
		self.allBinaryVars = []
	
	def getCost(self,i,j):
		return self.costMatrix[i][j];
	
	def objectiveFunc(self, nodes):
		s = "Minimize\n"
		for i in nodes:
			for j in nodes:
				if(i == j):
					continue
				for k in nodes:
					varName = "r{0}_{1}_{2}".format(i,j,k)
					#cost = random.randint(1, 20);
					cost = self.getCost(i-1,j-1)
					s += "+ {0} {1} ".format(cost,varName)
					self.allBinaryVars.append(varName)
				s +="\n"
		return s
	
	def inOutDegreeConstraints(self, nodes):
		#print len(nodes)
		s = ""
		for i in nodes:
			sumStr = ""	
			for j in nodes:
				if(i == j):
					continue				
				for k in nodes:
					varName = "r{0}_{1}_{2}".format(i,j,k)
					sumStr += "+ {0} ".format(varName)
				# for formatiing
				if(j == len(nodes)):
					sumStr = sumStr
				elif(i == len(nodes) and j == len(nodes) - 1):
					sumStr = sumStr
				else:
					sumStr += "\n"
										
			sumStr = sumStr + " = 1\n\n"
			s += sumStr	
		
		s += "\n"
		for j in nodes:
			sumStr = ""			
			for i in nodes:
				if(i == j):
					continue
				for k in nodes:
					varName = "r{0}_{1}_{2}".format(i,j,k)
					sumStr += "+ {0} ".format(varName)
				# for formatiing
				if(i == len(nodes)):
					sumStr = sumStr
				elif(j == len(nodes) and i == len(nodes) - 1):
					sumStr = sumStr
				else:
					sumStr += "\n"										
			sumStr = sumStr + " = 1\n\n"
			s += sumStr	
		
		return s
		
	def subToursEliminationConstraints(self, nodes):
		s ="\n"
		totalNodes = len(nodes)
		#Eq 32
		for i in range(2, totalNodes + 1):
			sum_ij_Str = ""
			sum_ji_Str = ""
			for j in range(1, totalNodes + 1):
				if(i == j ): continue
				
				for k in range(2, totalNodes + 1):
					r_ijk = "{2} r{0}_{1}_{2}".format(i,j,k)
					sum_ij_Str += "+ {0} ".format(r_ijk)
				sum_ij_Str += "\n"
				
				for k in range(1, totalNodes + 1):
					r_jik = "{2} r{0}_{1}_{2}".format(j,i,k)
					sum_ji_Str += "- {0} ".format(r_jik)
				
				if(j == totalNodes):
					sum_ji_Str = sum_ji_Str
				elif(i == totalNodes and j == totalNodes - 1):
					sum_ji_Str = sum_ji_Str
				else:
					sum_ji_Str += "\n"
				
			s += sum_ij_Str
			s += sum_ji_Str + " = 1"
			s += "\n\n"
		return s
	
	def getBounds(self, nodes):
		s ="Bounds\n"
		totalNodes = len(nodes)
		for i in range(2, totalNodes + 1):
			c = "2 <= u{0} <= {1}\n".format(i, totalNodes)
			s += c
		s += "\n"
		return s
	
	def getBinaries(self, nodes):
		s ="Binary\n"
		totalNodes = len(nodes)
		for binaryVar in self.allBinaryVars:
			s += binaryVar + "\n"
		return s	
		
	def generateEncoding(self,totalNodes):
		nodes = []
		self.allBinaryVars = []
		#print nodes
		for n in range(1, totalNodes + 1):
			nodes.append(n);
		#print nodes
		string = "";
		# Eq 28
		string += self.objectiveFunc(nodes) + "\n"
		string += "Subject To\n"
		# Eq 29 and 30
		string += self.inOutDegreeConstraints(nodes)
		# Eq 32
		string += self.subToursEliminationConstraints(nodes)
		#string += self.getBounds(nodes)		
		string += self.getBinaries(nodes)
		string += "End"
		return string
	
	def printCostMatrix(self, costMatrix):
		for i in range(len(costMatrix)):
			print costMatrix[i]
	
	def getCostMatrixFromFile(self, costMatrixFilePath):
		costMatrix = []
		self.numberOfNodes = 0;
		self.euclideanMaximumCost = 0;
		with open(costMatrixFilePath) as infile:
			for line in infile:			
				costs = line.split(',')
				if(len(costs) == 1):
					self.numberOfNodes = int(costs[0])
					continue
				intCosts = []
				for c in costs:
					cost = int(c)
					intCosts.append(int(c))
					if( cost > self.euclideanMaximumCost):
						self.euclideanMaximumCost = cost						
				costMatrix.append(intCosts)
				
		return costMatrix
	
	def deleteThanCreateFolder(self, path):
		if(os.path.exists(path)):
			shutil.rmtree(path)
		os.makedirs(path)
	
	def createIfNotExists(self, path):
		if(os.path.exists(path) == False):
			os.makedirs(path)
	
	def generate(self):
		
		print "It's exact same as FGG encoding of binary !!"
			
		self.benchMarkFolder = "../../../BenchMarks";
		costFilesPath = ['EuclideanCost.txt','OceanDataCost.txt','RandomCost.txt',
				'noisy_EuclideanCost/RandomSymmetric_EuclideanCostModified.txt',
				'noisy_EuclideanCost/AsymmetricRandomNoisy_50PercentEuclideanCostModified.txt',
				'noisy_EuclideanCost/AsymmetricNoisy_50PercentEuclideanCostModified.txt',
				'noisy_EuclideanCost/SymmetricNoisy_50PercentEuclideanCostModified.txt'
				]
		
		#costFilesPath = ['OceanDataCost.txt']
		
		nodeFolders  = os.listdir(self.benchMarkFolder)
		for nodeFolderName in nodeFolders:
			print "-------------"
			#if(nodeFolderName != "5_Nodes" and nodeFolderName != "4_Nodes"): continue;
			craftedFolderPath = self.benchMarkFolder + "/" + nodeFolderName
			if(os.path.isdir(craftedFolderPath)):
				setFolders  = os.listdir(craftedFolderPath)
				for setFolder in setFolders:
					#if(setFolder != "baseTest_set"): continue;
					setFolderPath = craftedFolderPath + "/" + setFolder
					encodingFolderPath = setFolderPath + "/encodings"
					self.createIfNotExists(encodingFolderPath)
					
					cplexEncodingFolderPath = setFolderPath + "/encodings/cplex"
					self.createIfNotExists(cplexEncodingFolderPath)
					
					cplexEncodingFolderPath = setFolderPath + "/encodings/cplex/mip"
					self.createIfNotExists(cplexEncodingFolderPath)
					
					cplexDLencodingFolderPath = setFolderPath + "/encodings/cplex/mip/fggEncoding"		
					self.deleteThanCreateFolder(cplexDLencodingFolderPath)					
					
					if(setFolder != "baseTest_set"): 
						if(os.path.exists(cplexDLencodingFolderPath)):
							shutil.rmtree(cplexDLencodingFolderPath)
						continue;
					
					costFilespath = setFolderPath + "/costFiles/";
					for costfile in costFilesPath:
						costFilePath = costFilespath + costfile
						if(os.path.exists(costFilePath)):
							self.costMatrix = self.getCostMatrixFromFile(costFilePath)
							#print costFilePath
							#self.printCostMatrix(self.costMatrix)
							encoding = self.generateEncoding(self.numberOfNodes)
							costFilename =  os.path.basename(costFilePath).split('.')[0]
							outputFilePath = cplexDLencodingFolderPath + "/cplex_mip_fgg_" + costFilename + ".lp"
							#print outputFilePath
							fileObj = open(outputFilePath,"w+")
							fileObj.write(encoding)
							fileObj.close()
						else:
							print "Not Exist: {0}".format(costFilePath)
				print nodeFolderName
				print "-------------"

if __name__ == '__main__':	
	gen = CplexBenchmarkGen()
	gen.generate()
	'''
	encoding = gen.generateEncoding(4)
	fileObj = open("out.lp","w+")
	fileObj.write(encoding)
	fileObj.close()
	'''
