import time
import os
import math
import random
import shutil

class CplexBenchmarkGen:
	def __init__(self):
		self.varMap = 0
	
	def getCost(self,i,j):
		return self.costMatrix[i][j];
	
	def objectiveFunc(self, nodes):
		s = "Minimize\n"
		for i in nodes:
			for j in nodes:
				if(i == j):
					continue
				varName = "x{0}_{1}".format(i,j)
				#cost = random.randint(1, 20);
				cost = self.getCost(i-1,j-1)
				s += "+ {0} {1} ".format(cost,varName)
			s +="\n"
		return s
	
	def inOutDegreeConstraints(self, nodes):
		s = ""
		for i in nodes:
			sumStr = ""
			for j in nodes:
				if(i == j):
					continue
				varName = "x{0}_{1}".format(i,j)
				sumStr += "+ {0} ".format(varName)
			s += sumStr + " = 1\n"
		
		s += "\n"
		for j in nodes:
			sumStr = ""
			for i in nodes:
				if(i == j):
					continue
				varName = "x{0}_{1}".format(i,j)
				sumStr += "+ {0} ".format(varName)
			s += sumStr + " = 1\n"	
		
		return s
		
	def subTours(self, nodes):
		s ="\n"
		totalNodes = len(nodes)
		for i in range(2, totalNodes + 1):
			for j in range(2, totalNodes + 1):
				if(i == j ): 
					continue
				varName = "x{0}_{1}".format(i,j)
				p = totalNodes -1
				# ui - uj + p xij <= p-1
				c = "u{0} - u{1} + {2} {3} <= {4}\n".format(i,j,p,varName,p - 1)
				s += c
			s += "\n"
		return s
	
	def getBounds(self, nodes):
		s ="Bounds\n"
		totalNodes = len(nodes)
		for i in range(2, totalNodes + 1):
			c = "2 <= u{0} <= {1}\n".format(i, totalNodes)
			s += c
		s += "\n"
		return s
	
	def getBinaries(self, nodes):
		s ="Binary\n"
		totalNodes = len(nodes)
		for i in nodes:
			for j in nodes:
				if(i == j):
					continue
				varName = "x{0}_{1}\n".format(i,j)
				s += varName
		return s	
		
	def generateEncoding(self,totalNodes):
		nodes = []
		for n in range(1, totalNodes + 1):
			nodes.append(n);
		string = "";	
		string += self.objectiveFunc(nodes) + "\n"
		string += "Subject To\n"
		string += self.inOutDegreeConstraints(nodes)
		string += self.subTours(nodes)
		string += self.getBounds(nodes)
		string += self.getBinaries(nodes)
		string += "End"
		return string
	
	def printCostMatrix(self, costMatrix):
		for i in range(len(costMatrix)):
			print costMatrix[i]
	
	def getCostMatrixFromFile(self, costMatrixFilePath):
		costMatrix = []
		self.numberOfNodes = 0;
		self.euclideanMaximumCost = 0;
		with open(costMatrixFilePath) as infile:
			for line in infile:			
				costs = line.split(',')
				if(len(costs) == 1):
					self.numberOfNodes = int(costs[0])
					continue
				intCosts = []
				for c in costs:
					cost = int(c)
					intCosts.append(int(c))
					if( cost > self.euclideanMaximumCost):
						self.euclideanMaximumCost = cost						
				costMatrix.append(intCosts)
				
		return costMatrix
	
	def deleteThanCreateFolder(self, path):
		if(os.path.exists(path)):
			shutil.rmtree(path)
		os.makedirs(path)
	
	def createIfNotExists(self, path):
		if(os.path.exists(path) == False):
			os.makedirs(path)
	
	def generate(self):
		
		self.benchMarkFolder = "../../../BenchMarks";
		costFilesPath = ['EuclideanCost.txt','OceanDataCost.txt','RandomCost.txt',
				'noisy_EuclideanCost/RandomSymmetric_EuclideanCostModified.txt',
				'noisy_EuclideanCost/AsymmetricRandomNoisy_50PercentEuclideanCostModified.txt',
				'noisy_EuclideanCost/AsymmetricNoisy_50PercentEuclideanCostModified.txt',
				'noisy_EuclideanCost/SymmetricNoisy_50PercentEuclideanCostModified.txt'
				]
		
		nodeFolders  = os.listdir(self.benchMarkFolder)
		for nodeFolderName in nodeFolders:
			print "-------------"
			#if(nodeFolderName != "4_Nodes"): continue;
			craftedFolderPath = self.benchMarkFolder + "/" + nodeFolderName
			if(os.path.isdir(craftedFolderPath)):
				print nodeFolderName
				setFolders  = os.listdir(craftedFolderPath)
				for setFolder in setFolders:
					#if(setFolder != "baseTest_set"): continue;
					setFolderPath = craftedFolderPath + "/" + setFolder
					encodingFolderPath = setFolderPath + "/encodings"
					self.createIfNotExists(encodingFolderPath)
					
					cplexEncodingFolderPath = setFolderPath + "/encodings/cplex"
					self.createIfNotExists(cplexEncodingFolderPath)
					
					cplexEncodingFolderPath = setFolderPath + "/encodings/cplex/mip"
					self.createIfNotExists(cplexEncodingFolderPath)
					
					cplexEncodingFolderPath = setFolderPath + "/encodings/cplex/mip/mtzEncoding"
					self.deleteThanCreateFolder(cplexEncodingFolderPath)
					
					costFilespath = setFolderPath + "/costFiles/";
					for costfile in costFilesPath:
						costFilePath = costFilespath + costfile
						if(os.path.exists(costFilePath)):
							self.costMatrix = self.getCostMatrixFromFile(costFilePath)
							#print costFilePath
							#self.printCostMatrix(self.costMatrix)
							encoding = self.generateEncoding(self.numberOfNodes)
							costFilename =  os.path.basename(costFilePath).split('.')[0]
							outputFilePath = cplexEncodingFolderPath + "/cplex_mip_mtz_" + costFilename + ".lp"
							#print outputFilePath
							fileObj = open(outputFilePath,"w+")
							fileObj.write(encoding)
							fileObj.close()
						else:
							print "Not Exist: {0}".format(costFilePath)
			print "-------------"

if __name__ == '__main__':	
	gen = CplexBenchmarkGen()
	gen.generate()
	'''
	encoding = gen.generateEncoding(4)
	fileObj = open("out.lp","w+")
	fileObj.write(encoding)
	fileObj.close()
	'''
