import random
import timeit
import datetime
import time
import os
import math
from optparse import OptionParser
import shutil
	

class PBencodingGenerator:
	
	def __init__(self):
		self.variableCounter = 0;
		self.constraintCounter = 0;
		self.allBinaryVars = []

	def getaFileobject(self, outputFilePath):
		filename = outputFilePath
		if os.path.exists(filename):
			os.remove(filename)
		fo = open(filename, "a")
		return fo
	
	def closeFileobject(self, fo):
		fo.close()
		
	def getCostMatrixFromFile(self, costMatrixFilePath):
		costMatrix = []
		self.numberOfNodes = 0;
		self.euclideanMaximumCost = 0;
		with open(costMatrixFilePath) as infile:
			for line in infile:			
				costs = line.split(',')
				if(len(costs) == 1):
					self.numberOfNodes = int(costs[0])
					continue
				intCosts = []
				for c in costs:
					cost = int(c)
					intCosts.append(int(c))
					if( cost > self.euclideanMaximumCost):
						self.euclideanMaximumCost = cost						
				costMatrix.append(intCosts)
				
		return costMatrix
	
	def getEdgekey(self, i, j):
		return "edge(" + str(i) + "," + str(j) + ")"
	
	def getNewVar(self):
		self.variableCounter += 1
		newvar = "x" + str(self.variableCounter)
		self.allBinaryVars.append(newvar)
		return newvar
	
	def generateObjectiveFunction(self, costMatrix, setFolderPath):
		string = "";
		self.edge_variable_map = {}
		
		edgeVarMapFilePath = setFolderPath + "/" + "EdgeVarMapFile.map";
		if(os.path.exists(edgeVarMapFilePath)):
			os.remove(edgeVarMapFilePath)		
		edgeVarMapFile = self.getaFileobject(edgeVarMapFilePath)

		for i in range(1, len(costMatrix) + 1):
			row = costMatrix[i-1]
			for j in range(1, len(row) + 1):
				if (i == j): continue
				edgeKey = self.getEdgekey(i,j)
				edgeVar = self.getNewVar()
				self.edge_variable_map[edgeKey] = edgeVar
				cost = str(row[j-1])
				string += "+" + cost + " " + edgeVar + " "
				evmStr = edgeKey + " : " + self.edge_variable_map[edgeKey] + "\n"
				edgeVarMapFile.write(evmStr)
			if(i != len(costMatrix)):
				string += "\n"
		
		edgeVarMapFile.close()
		
		return string + ";"
	
	def inOutDegreeConstraints(self, numberOfNodes):
		PBString = ""		  
		#"Out going edge from i"
		for i in range(1, numberOfNodes + 1):
			for j in range(1, numberOfNodes + 1):				
				if(i != j):
					edgekey = self.getEdgekey(i,j)
					edgeVar = self.edge_variable_map[edgekey]
					PBString += "+" + "1 " + edgeVar + " "
			PBString += "= 1 ;\n"
			self.constraintCounter += 1;

		
		PBString += "\n"
		#"In coming edge at j"
		for j in range(1, numberOfNodes + 1):
			for i in range(1, numberOfNodes + 1):
				if(i != j):
					edgekey = self.getEdgekey(i,j)
					edgeVar = self.edge_variable_map[edgekey]
					PBString += "+" + "1 " + edgeVar + " " 
			PBString += "= 1 ;\n"
			self.constraintCounter += 1;
		return PBString
	
	def getUvariable(self, nBits, sign = "+"):
		u = ""
		for i in range(0, len(nBits)):
			d = int(math.pow( 2, i ))
			u += sign + str(d) + " " + nBits[i] + " "
		return u
	
	def getNbitsVar(self, n):
		var = []
		for i in range(0, n ):
			v = self.getNewVar()
			var.append(v)
		return var
	
	def subToursEliminationConstraints(self, totalNodes):
		
		subTourConstraints = "\n"
		boundConstraints = "\n"	
		self.uVar_bits_maps = {}
		bitCount = len(bin(totalNodes)[2:])
		
		for i in range(2, totalNodes + 1):
			
			varkey = "u" + str(i)
			#subTourConstraints += "*" + varkey + "\n"
			uiBits = self.getNbitsVar(bitCount)
			self.uVar_bits_maps[varkey] = uiBits
			ui = self.getUvariable(uiBits)
			
			uiRangeConstraint =  ui + " >= 2 ;\n"
			boundConstraints +=  uiRangeConstraint
			self.constraintCounter += 1
			
			uiRangeConstraint =  ui + " <= {0} ;\n".format(totalNodes)
			boundConstraints +=  uiRangeConstraint
			self.constraintCounter += 1
			
			boundConstraints += "\n"
		
		for i in range(2, totalNodes + 1):
			uiVarkey = "u" + str(i)
			uiBits = self.uVar_bits_maps[uiVarkey]
			ui = self.getUvariable(uiBits)
				   
			for j  in range(2, totalNodes + 1):
				if (i == j): continue		
				
				ujVarkey = "u" + str(j)
				ujBits = self.uVar_bits_maps[ujVarkey]
				negUj = self.getUvariable(ujBits,"-")
								
				edgekey = self.getEdgekey(i,j)
				varName = self.edge_variable_map[edgekey]				
				
				p = totalNodes -1
				# ui - uj + p xij <= p-1
				#subTourConstraints += "*u{0} - u{1} + {2} {3} <= {4}\n".format(i,j,p,varName,p - 1)	
				subTourConstraints +=  ui + negUj + "+{0} {1} <= {2} ;\n".format(p,varName,p-1)
				self.constraintCounter += 1
			
			subTourConstraints += "\n"
				
		return (subTourConstraints, boundConstraints)
	
	def getAllBinaryVars(self):
		string = ""
		for binaryVar in self.allBinaryVars:
			string += binaryVar + "\n"
		return string
	
	def generateEncoding(self,totalNodes,setFolderPath):
		
		self.variableCounter = 0;
		self.constraintCounter = 0;
		self.allBinaryVars = []
		
		#(costMatrix, numberOfNodes) = self.getCostMatrixFromFile(options.costMatrixFilePath)
		costMatrix = self.costMatrix
		numberOfNodes = totalNodes
		
		objectiveFunction = self.generateObjectiveFunction(costMatrix, setFolderPath)
		in_out_edge_contraint = self.inOutDegreeConstraints(numberOfNodes)
		
		PBstring = "min: \n";
		PBstring += objectiveFunction + "\n"
		PBstring += "\n" + in_out_edge_contraint
		(subTourConstraints, boundConstraints) = self.subToursEliminationConstraints(numberOfNodes)
		PBstring += subTourConstraints
		PBstring += boundConstraints
		fileHeader = "* #variable= {0} #constraint= {1}\n".format(self.variableCounter, self.constraintCounter)
		PBstring = fileHeader + PBstring

		return PBstring

	def optionParse(self):
		parser = OptionParser()
		parser.add_option("--costMatrixFilePath", dest="costMatrixFilePath", help="Cost Matrix File path", type="string", default  = 'EuclideanDistanceCostmatrix.txt' )
		parser.add_option("--outputFilePath", dest="outputFilePath", help="output file path", type="string" , default  = 'PBEncoding_From_EuclideanDistanceCostmatrix.opb')
		return parser.parse_args()
	
	def creatFolder(self, path):
		if(os.path.exists(path)):
			shutil.rmtree(path)
		os.makedirs(path)
	
	def deleteThanCreateFolder(self, path):
		if(os.path.exists(path)):
			shutil.rmtree(path)
		os.makedirs(path)
	
	def createIfNotExists(self, path):
		if(os.path.exists(path) == False):
			os.makedirs(path)
			
	def generate(self):
		
		self.benchMarkFolder = "../../../BenchMarks";
		costFilesPath = ['EuclideanCost.txt','OceanDataCost.txt','RandomCost.txt',
				'noisy_EuclideanCost/RandomSymmetric_EuclideanCostModified.txt',
				'noisy_EuclideanCost/AsymmetricRandomNoisy_50PercentEuclideanCostModified.txt',
				'noisy_EuclideanCost/AsymmetricNoisy_50PercentEuclideanCostModified.txt',
				'noisy_EuclideanCost/SymmetricNoisy_50PercentEuclideanCostModified.txt'
				]		
		nodeFolders  = os.listdir(self.benchMarkFolder)
		
		for nodeFolderName in nodeFolders:
			
			#if(nodeFolderName != "4_Nodes"): continue;
			craftedFolderPath = self.benchMarkFolder + "/" + nodeFolderName
			
			if(os.path.isdir(craftedFolderPath)):
				
				setFolders  = os.listdir(craftedFolderPath)
				print "-------------"
				for setFolder in setFolders:					
					#if(setFolder != "baseTest_set"): continue;
					
					setFolderPath = craftedFolderPath + "/" + setFolder
					#print setFolderPath
					encodingFolderPath = setFolderPath + "/encodings"
					self.createIfNotExists(encodingFolderPath)
					
					pboEncodingFolderPath = setFolderPath + "/encodings/sat4j_scip"
					self.createIfNotExists(pboEncodingFolderPath)
										
					pboEncodingFolderPath = setFolderPath + "/encodings/sat4j_scip/mtzEncoding"
					self.deleteThanCreateFolder(pboEncodingFolderPath)
					
					costFilespath = setFolderPath + "/costFiles/";
					for costfile in costFilesPath:
						costFilePath = costFilespath + costfile
						if(os.path.exists(costFilePath)):
							self.costMatrix = self.getCostMatrixFromFile(costFilePath)
							#print costFilePath
							#self.printCostMatrix(self.costMatrix)
							encoding = self.generateEncoding(self.numberOfNodes, setFolderPath)
							costFilename =  os.path.basename(costFilePath).split('.')[0]
							outputFilePath = pboEncodingFolderPath + "/opb_mtz_" + costFilename + ".opb"
							#print outputFilePath
							#print outputFilePath
							fileObj = open(outputFilePath,"w+")
							fileObj.write(encoding)
							fileObj.close()
						else:
							print "Not Exist: {0}".format(costFilePath)
				print "-------------"
			

if __name__ == '__main__':
	
	generator = PBencodingGenerator()
	generator.generate()
	
	

