from z3 import *
import random
import timeit
import datetime
import time
import os
import math
import random
import shutil

class IncrementalSMTEncoding:
	def __init__(self):
		self.varMap = 0
	
	def validateTour(selectedEdges, startNode, traversedNode, traversedEdgeCounter, totalNode):
		node2 = None
		for index in range(len(selectedEdges)):
			var = selectedEdges[index]
			varName = str(var)[1:]
			nodes = varName.split('_');
			node1 = int(nodes[0])	   
			if(node1 == startNode):
				traversedEdgeCounter += 1
				node2 = int(nodes[1])
				if( node1 not in traversedNode):
					traversedNode.append(node1)
				if( node2 in traversedNode and traversedEdgeCounter < totalNode):
					 traversedNode.append(node2)
					 print "\nINVALID TOUR at " + str(node1) + "->" + str(node2)
					 return False
				else:
					traversedNode.append(node2)
				#print node1, node2		 
				break
		if(traversedEdgeCounter == totalNode):
			print str(node1) + " " + str(node2) + ". Stop calling. Exit !"
			print "\nVALID TOUR:"
			return
		if(node2 != None):
			print str(node1) + " " + str(node2) + ". Calling for " + str(node2)  +", counter " + str(traversedEdgeCounter)
			validateTour(selectedEdges,node2, traversedNode, traversedEdgeCounter, totalNode)
		
		return True
	
	def getCostMatrixFromFile(self, costMatrixFilePath):
		costMatrix = []
		self.numberOfNodes = 0;
		self.euclideanMaximumCost = 0;
		with open(costMatrixFilePath) as infile:
			for line in infile:			
				costs = line.split(',')
				if(len(costs) == 1):
					self.numberOfNodes = int(costs[0])
					continue
				intCosts = []
				for c in costs:
					cost = int(c)
					intCosts.append(int(c))
					if( cost > self.euclideanMaximumCost):
						self.euclideanMaximumCost = cost						
				costMatrix.append(intCosts)
				
		return costMatrix
	
	def printCostMatrix(self, costMatrix):
		for i in range(len(costMatrix)):
			print costMatrix[i]
	
	def getEdgeKey(self,i,j):
		return "e{0}_{1}".format(i,j)
	
	def getEdgeVariablesMap(self,):
		edge_var_maps = {}
		for i in range(1, self.numberOfNodes + 1):
			for j in range(1, self.numberOfNodes + 1):
				if(i != j):
					var = self.getEdgeKey(i,j)
					z3IntVar = Int(var)
					edge_var_maps[var] = z3IntVar
		return edge_var_maps
	
	def inOutDegreeConstraints(self, nodes, z3Solver):

		for i in nodes:
			outgoingEdgeSum = []			
			for j in nodes:
				if(i == j):
					continue
				edgeKey = self.getEdgeKey(i,j)
				z3IntEdgeVar = self.edge_var_maps[edgeKey]
				outgoingEdgeSum.append(z3IntEdgeVar)
				
			z3Solver.add(Sum(outgoingEdgeSum) == 1)
		z3Solver.add(True)
			
		for j in nodes:
			incomingEdgeSum = []
			for i in nodes:
				if(i == j):
					continue
				edgeKey = self.getEdgeKey(i,j)
				z3IntEdgeVar = self.edge_var_maps[edgeKey]
				incomingEdgeSum.append(z3IntEdgeVar)				
			
			z3Solver.add(Sum(incomingEdgeSum) == 1)
		z3Solver.add(True)
	
	def subTours(self, nodes, z3Solver):

		totalNodes = len(nodes)
		for i in range(2, totalNodes + 1):
			for j in range(2, totalNodes + 1):
				if(i == j ): 
					continue
				edgeKey = self.getEdgeKey(i,j)
				z3IntEdgeVar = self.edge_var_maps[edgeKey]
				
				ui = Int("u{0}".format(i))
				uj = Int("u{0}".format(j))
				z3Solver.add( And(ui >= 2, ui <= totalNodes))
				z3Solver.add( And(uj >= 2, uj <= totalNodes))
								
				p = totalNodes -1
				# ui - uj + p xij <= p-1
				#c = "u{0} - u{1} + {2} {3} <= {4}\n".format(i,j,p,varName,p - 1)
				cons = ui - uj + p * z3IntEdgeVar <= p - 1
				z3Solver.add(cons)
			z3Solver.add(True)
				
	def objectiveFunc(self,nodes,z3Solver):
		#z3IntEgdeVariables = self.edge_var_maps.values()
		#return Sum(z3IntEgdeVariables)
		sumElment = []
		for i in nodes:			
			for j in nodes:
				if(i == j):
					continue
				edgeKey = self.getEdgeKey(i,j)
				z3IntEdgeVar = self.edge_var_maps[edgeKey]
				cost = self.costMatrix[i-1][j-1]
				sumElment.append(cost * z3IntEdgeVar)
				z3Solver.add(Or(z3IntEdgeVar == 0, z3IntEdgeVar == 1))
		return Sum(sumElment)
	
	def getSelectedEdgeFromModel(self,model):
		selectedEdges = []
		for edgeKey in  self.edge_var_maps:
			z3IntEdgeVar = self.edge_var_maps[edgeKey]
			val = model.evaluate(z3IntEdgeVar)			
			if(int(val.as_string()) == 1):
				selectedEdges.append(edgeKey)
		#print "\nModel: "
		#print selectedEdges
		#print "\n"
		return selectedEdges
	
	def getCostFromModel(self, model):
		tCost = 0;
		for edgeKey in model:
			edgeKey = edgeKey[1:]
			parts = edgeKey.split("_")
			i = int(parts[0])
			j = int(parts[1])
			cost = self.costMatrix[i-1][j-1]
			tCost += cost
		#print tCost
		return tCost
	
	def solveWithGivenBound(self, z3Solver):
		if z3Solver.check() == sat:
			model = z3Solver.model()
			selectedEdges = self.getSelectedEdgeFromModel(model)
			return self.getCostFromModel(selectedEdges)
		else:
			print "\nUNSAT"
		return -1
		
	
	def start(self):
		self.benchMarkFolder = "../../BenchMarks";
		costFilePath = self.benchMarkFolder + "/15_Nodes/baseTest_set/costFiles/EuclideanCost.txt";
		
		self.costMatrix = self.getCostMatrixFromFile(costFilePath)
		#self.printCostMatrix(self.costMatrix)
		
		nodes = []
		for n in range(1, self.numberOfNodes  + 1):
			nodes.append(n);
		
		z3Solver = Solver()
		
		self.edge_var_maps = self.getEdgeVariablesMap()
		self.inOutDegreeConstraints(nodes,z3Solver)
		self.subTours(nodes,z3Solver)
		objFunc = self.objectiveFunc(nodes,z3Solver)
		#bound = 16544
		z3Solver.add(objFunc >= 0)
		fisrtCost = self.solveWithGivenBound(z3Solver)
		currentCost = fisrtCost
		print "First cost: {0}".format(fisrtCost)
		
		i =0
		while (1):
			bound = currentCost - 1
			z3Solver.push()
			z3Solver.add(objFunc <= bound)
			currentCost = self.solveWithGivenBound(z3Solver)
			print "{0} : {1} -> {2}" .format(i, bound, currentCost)
			z3Solver.pop()
			if(currentCost == -1): break
			bound = currentCost
			i = i +1
		
		
if __name__ == '__main__':	
	gen = IncrementalSMTEncoding()
	gen.start()

